@@ -146,12 +146,12 @@
 //! the device.  If a device contains multiple keys, and they can be
 //! registered, or available independent of the other keys, then the
 //! backend must model the keys as separate devices.
-use std::cell::OnceCell;
 
 use sequoia_openpgp as openpgp;
 use openpgp::Fingerprint;
 use openpgp::KeyID;
 use openpgp::Result;
+use openpgp::Cert;
 use openpgp::crypto::Password;
 use openpgp::crypto::SessionKey;
 use openpgp::crypto::mpi;
@@ -160,6 +160,21 @@ use openpgp::types::HashAlgorithm;
 use openpgp::types::PublicKeyAlgorithm;
 use openpgp::types::SymmetricAlgorithm;
 
+/// A testing framework for backends that implement this interface.
+///
+/// Define a couple of functions and then use
+/// [`sequoia_keystore_backend::generate_tests!`] to generate a number
+/// of standard tests.
+pub mod test_framework;
+
+pub mod utils;
+
+mod protection;
+pub use protection::Protection;
+
+mod password_source;
+pub use password_source::PasswordSource;
+
 /// Errors used in this crate.
 ///
 /// Note: This enum cannot be exhaustively matched to allow for future
@@ -215,12 +230,63 @@ pub enum Error {
     /// unlocked.
     #[error("Device is already unlocked: {0}")]
     AlreadyUnlocked(String),
+
+    /// The backend doesn't support importing keys using this interface.
+    ///
+    /// This error is returned by [`Backend::import`] if the backend
+    /// requires backend-specific information in order to import a
+    /// key, which is not provided via this interface.  For instance,
+    /// when importing a key to a smartcard, the user needs to specify
+    /// the smartcard, and the slot to use on the smartcard.  Rather
+    /// than try and model these parameters using this generic
+    /// interface, backends should just have their own tools.  The
+    /// text should be a hint that is displayed to the user describing
+    /// how to find the tool.
+    #[error("Can't import keys into the backend: {}",
+            .0.as_deref().unwrap_or("use another tool to import keys \
+                                     into this backend"))]
+    ExternalImportRequired(Option<String>),
+
+    /// The key doesn't support inline passwords.
+    ///
+    /// Password entry is taken care of by the device managing the
+    /// key.  For instance, the password may be obtained using an
+    /// external PIN pad.
+    #[error("Can't unlock using an inline password{}{}",
+            .0.as_ref().map(|_| ": ").unwrap_or(""),
+            .0.as_deref().map(|msg| msg).unwrap_or(""))]
+    NoInlinePassword(Option<String>),
+
+    /// The key doesn't support getting passwords.
+    ///
+    /// The password must be provided inline.  The password cannot be
+    /// obtained using something like an external PIN pad.
+    #[error("External password entry is not supported{}{}",
+            .0.as_ref().map(|_| ": ").unwrap_or(""),
+            .0.as_deref().map(|msg| msg).unwrap_or(""))]
+    NoExternalPassword(Option<String>),
+}
+
+/// The result of an import operation.
+///
+/// This is returned by [`Backend::import`].
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub enum ImportStatus {
+    /// The imported object is new.
+    New,
+
+    /// The imported object updated an existing object.
+    Updated,
+
+    /// The object already existed, and is unchanged.
+    Unchanged,
 }
 
 /// The backend interface for the sequoia key store.
 ///
 /// This is the the interface that each device driver needs to
 /// implement.
+#[async_trait::async_trait]
 pub trait Backend: Send {
     /// Returns the backend's identifier.
     ///
@@ -255,7 +321,7 @@ pub trait Backend: Send {
     ///
     /// In the terminology of [`DeviceHandle`], this should look for
     /// devices that are available or registered.
-    fn scan(&mut self) -> Result<()> {
+    async fn scan(&mut self) -> Result<()> {
         Err(Error::OperationNotSupported("Backend::scan".into()).into())
     }
 
@@ -266,7 +332,7 @@ pub trait Backend: Send {
     ///
     /// In the terminology of [`DeviceHandle`], it returns devices
     /// that are available or are registered.
-    fn list<'a>(&'a self)
+    async fn list<'a>(&'a self)
         -> Box<dyn Iterator<Item=Box<dyn DeviceHandle + Send + Sync + 'a>>
                + Send + Sync + 'a>;
 
@@ -276,7 +342,7 @@ pub trait Backend: Send {
     ///
     /// If the device is not available or not registered, then this
     /// should return [`Error::NotFound`].
-    fn find_device<'a>(&self, id: &str)
+    async fn find_device<'a>(&self, id: &str)
         -> Result<Box<dyn DeviceHandle + Send + Sync + 'a>>;
 
     /// Returns a handle for the specified key.
@@ -285,8 +351,24 @@ pub trait Backend: Send {
     ///
     /// If the key is not available or not registered, then this
     /// should return [`Error::NotFound`].
-    fn find_key<'a>(&self, id: &str)
+    async fn find_key<'a>(&self, id: &str)
         -> Result<Box<dyn KeyHandle + Send + Sync + 'a>>;
+
+    /// Imports the keys in the cert.
+    ///
+    /// `cert` is a TSK.  Any keys without secret key material are
+    /// silently ignored.
+    ///
+    /// Returns each key's import status, and a capability to the key.
+    ///
+    /// If the TSK doesn't include any secret keys, then an empty list
+    /// is returned.
+    ///
+    /// Some backends require additional information to import a key.
+    /// These backends should [`Error::ExternalImportRequired`], and
+    /// indicate how a user might import a key to this backend.
+    async fn import<'a>(&self, cert: Cert)
+        -> Result<Vec<(ImportStatus, Box<dyn KeyHandle + Send + Sync + 'a>)>>;
 }
 
 /// A device that contains zero or more keys.
@@ -397,7 +479,7 @@ pub trait Backend: Send {
 /// ```
 ///
 /// A device is only ever instantiated if it is available or registered.
-#[blanket::blanket(derive(Box, Mut))]
+#[async_trait::async_trait]
 pub trait DeviceHandle {
     /// Returns the device's id.
     ///
@@ -409,33 +491,33 @@ pub trait DeviceHandle {
     ///
     /// The description is a human readable string, e.g., "GnuK with
     /// certificate FINGERPRINT".
-    fn description(&self) -> String {
+    async fn description(&self) -> String {
         self.id()
     }
 
     /// Sets the device's human readable description.
     ///
     /// This should fail if the device is not registered.
-    fn set_description(&self, _description: String) -> Result<()> {
+    async fn set_description(&self, _description: String) -> Result<()> {
         Err(Error::Unregistered(self.id()).into())
     }
 
     /// Returns whether the device is available.
     ///
     /// A device is available if it is plugged-in.
-    fn available(&self) -> bool;
+    async fn available(&self) -> bool;
 
     /// Returns whether the device is configured.
     ///
     /// A device is configured if the keys can be used without further
     /// configuration.
-    fn configured(&self) -> bool;
+    async fn configured(&self) -> bool;
 
     /// Returns whether the device is registered.
     ///
     /// A device is registered if the device is locally memorized.  In
     /// this case, the user may be prompted to insert the device.
-    fn registered(&self) -> bool;
+    async fn registered(&self) -> bool;
 
     /// Registers the device.
     ///
@@ -443,7 +525,7 @@ pub trait DeviceHandle {
     /// memorize the device and should return it during a scan even if
     /// it is not available.  If the device is already registered,
     /// this should silently succeed.
-    fn register(&mut self) -> Result<()> {
+    async fn register(&mut self) -> Result<()> {
         Err(Error::OperationNotSupported("DeviceHandle::register".into()).into())
     }
 
@@ -460,7 +542,7 @@ pub trait DeviceHandle {
     /// Note: devices that are not available can only be registered
     /// using a backend-specific tool.  For instance, a device
     /// accessible via an ssh tunnel is never available.
-    fn unregister(&mut self) -> Result<()> {
+    async fn unregister(&mut self) -> Result<()> {
         Err(Error::OperationNotSupported("DeviceHandle::unregister".into()).into())
     }
 
@@ -472,7 +554,7 @@ pub trait DeviceHandle {
     /// enumerated.  For instance, if soft keys are stored in a
     /// database and the database is encrypted, it may be necessary to
     /// supply a password to decrypt the database.
-    fn unlock(&mut self, _password: &Password) -> Result<()> {
+    async fn unlock(&mut self, _password: &Password) -> Result<()> {
         Err(Error::AlreadyUnlocked("DeviceHandle::unlock".into()).into())
     }
 
@@ -484,7 +566,7 @@ pub trait DeviceHandle {
     /// deinitialized, it MAY be deinitialized lazily if doing so
     /// cannot result in a user-visible error.  For instance, if the
     /// device uses an ssh tunnel, the ssh tunnel may be closed later.
-    fn lock(&mut self) -> Result<()> {
+    async fn lock(&mut self) -> Result<()> {
         Ok(())
     }
 
@@ -492,7 +574,7 @@ pub trait DeviceHandle {
     ///
     /// Returns the keys on the device.  If the device is not usable,
     /// then this should return [`Error::Unconfigured`].
-    fn list<'a>(&'a self)
+    async fn list<'a>(&'a self)
         -> Box<dyn Iterator<Item=Box<dyn KeyHandle + Send + Sync + 'a>>
                + Send + Sync + 'a>;
 }
@@ -501,7 +583,7 @@ pub trait DeviceHandle {
 ///
 /// A key may or may not be available.  This is a function of the
 /// device.
-#[blanket::blanket(derive(Box, Mut))]
+#[async_trait::async_trait]
 pub trait KeyHandle {
     /// Returns the key's id.
     ///
@@ -512,36 +594,6 @@ pub trait KeyHandle {
     /// 5 #217813388320."
     fn id(&self) -> String;
 
-    /// Returns whether the key is available.
-    fn available(&self) -> bool;
-
-    /// Returns whether the key is locked.
-    fn locked(&self) -> bool;
-
-    /// Returns whether the key is decryption capable.
-    fn decryption_capable(&self) -> bool;
-
-    /// Returns whether the key is signing capable.
-    fn signing_capable(&self) -> bool;
-
-    /// Unlocks a key.
-    ///
-    /// A key is typically unlocked by providing a password or pin.
-    /// Not all keys are locked.  If the key is not available, this
-    /// should attempt to connect to the device.  If the device is not
-    /// available or cannot be initialized, then this should fail.
-    fn unlock(&mut self, _password: &Password) -> Result<()> {
-        Err(Error::AlreadyUnlocked("KeyHandle::unlock".into()).into())
-    }
-
-    /// Lock a key.
-    ///
-    /// Relocks the key.  This usually causes the backend to forget the
-    /// key's password.
-    fn lock(&mut self) -> Result<()> {
-        Ok(())
-    }
-
     /// Returns the key's fingerprint.
     fn fingerprint(&self) -> Fingerprint;
 
@@ -550,17 +602,69 @@ pub trait KeyHandle {
         KeyID::from(self.fingerprint())
     }
 
+    /// Returns the key's device.
+    async fn device<'a>(&self) -> Box<dyn DeviceHandle + Send + Sync + 'a>;
+
+    /// Returns whether the key is available.
+    async fn available(&self) -> bool;
+
+    /// Returns whether the key is locked.
+    async fn locked(&self) -> Protection;
+
+    /// Returns how the password is obtained.
+    ///
+    /// This is similar to, but not identical to
+    /// [`KeyHandle::locked`].  This function indicates how the
+    /// password must be provided independency of the current
+    /// protection.
+    async fn password_source(&self) -> PasswordSource;
+
+    /// Returns whether the key is decryption capable.
+    async fn decryption_capable(&self) -> bool;
+
+    /// Returns whether the key is signing capable.
+    async fn signing_capable(&self) -> bool;
+
+    /// Unlocks a key.
+    ///
+    /// A key is typically unlocked by providing a password or pin.
+    /// Not all keys are locked.  If the key is not available, this
+    /// should attempt to connect to the device.  If the device is not
+    /// available or cannot be initialized, then this should fail.
+    ///
+    /// If `password` is `Some` and [`KeyHandle::password_source`]
+    /// indicates that the password cannot be provided inline, then
+    /// the backend must return [`Error::NoInlinePassword`].  Likewise,
+    /// if `password` is `None`, and [`KeyHandle::password_source`]
+    /// does not indicate that the user can be prompted for the
+    /// password ([`PasswordSource::ExternalOnDemand`]), then the
+    /// backend must return [`Error::NoExternalPassword`].
+    ///
+    /// If the key is already unlocked, this returns
+    /// [`Error::AlreadyUnlocked`].
+    async fn unlock(&mut self, _password: Option<&Password>) -> Result<()> {
+        Err(Error::AlreadyUnlocked("KeyHandle::unlock".into()).into())
+    }
+
+    /// Lock a key.
+    ///
+    /// Relocks the key.  This usually causes the backend to forget the
+    /// key's password.
+    async fn lock(&mut self) -> Result<()> {
+        Ok(())
+    }
+
     /// Returns the corresponding public key.
     ///
     /// The backend SHOULD ensure that the secret key material is
     /// removed.
-    fn public_key(&self)
+    async fn public_key(&self)
         -> openpgp::packet::Key<openpgp::packet::key::PublicParts,
                                 openpgp::packet::key::UnspecifiedRole>;
 
     /// Decrypts a PKESK.
     ///
-    fn decrypt_pkesk(&mut self, pkesk: &PKESK)
+    async fn decrypt_pkesk(&mut self, pkesk: &PKESK)
         -> Option<(SymmetricAlgorithm, SessionKey)>
     {
         // We want to use `PKESK::decrypt`.  For that, we need
@@ -568,40 +672,66 @@ pub trait KeyHandle {
         // could implement `Decryptor` for `KeyHandle`, but then that
         // is part of our public API.  Instead, we do a bit of
         // acrobatics here.
-        struct Decryptor<T>
-            where T: KeyHandle
+        struct Decryptor<'a, T>
+            where T: KeyHandle + ?Sized
         {
-            slf: T,
-            pk: OnceCell<
-                    openpgp::packet::Key<
-                            openpgp::packet::key::PublicParts,
-                            openpgp::packet::key::UnspecifiedRole>>,
+            slf: &'a mut T,
+            pk: openpgp::packet::Key<
+                    openpgp::packet::key::PublicParts,
+                    openpgp::packet::key::UnspecifiedRole>,
         }
 
-        impl<T> openpgp::crypto::Decryptor for Decryptor<T>
-            where T: KeyHandle
+        impl<'a, T> openpgp::crypto::Decryptor for Decryptor<'a, T>
+            where T: KeyHandle + ?Sized + Send
         {
             fn public(&self)
                 -> &openpgp::packet::Key<
                     openpgp::packet::key::PublicParts,
                     openpgp::packet::key::UnspecifiedRole>
             {
-                self.pk.get_or_init(|| self.slf.public_key())
+                &self.pk
             }
 
             fn decrypt(&mut self, ciphertext: &mpi::Ciphertext,
                        plaintext_len: Option<usize>)
                 -> Result<SessionKey>
             {
-                self.slf.decrypt_ciphertext(ciphertext, plaintext_len)
+                // We know that pkesk.decrypt is running on a separate
+                // thread.  So it is safe to create a new runtime on
+                // this thread.
+                let rt = tokio::runtime::Runtime::new()?;
+
+                rt.block_on(async {
+                    self.slf.decrypt_ciphertext(ciphertext, plaintext_len).await
+                })
             }
         }
+
+        let pk = self.public_key();
+        let pk = pk.await;
+
         let mut decryptor = Decryptor {
             slf: self,
-            pk: OnceCell::new(),
+            pk: pk,
         };
 
-        pkesk.decrypt(&mut decryptor, None)
+        // To avoid blocking the current thread, we run the sync
+        // function `pkesk.decrypt` on a separate thread.  When it is
+        // done, it results the result via a one shot channel, which
+        // we can asynchronously wait on.
+        let (sender, receiver) = futures::channel::oneshot::channel::<_>();
+
+        std::thread::scope(|s| {
+            s.spawn(move || {
+                sender.send(pkesk.decrypt(&mut decryptor, None))
+            });
+        });
+
+        match receiver.await {
+            Ok(Some(result)) => Some(result),
+            Ok(None) => None,
+            Err(_) => None,
+        }
     }
 
     /// Decrypts a ciphertext.
@@ -610,14 +740,48 @@ pub trait KeyHandle {
     /// [`sequoia_openpgp::crypto::Decryptor::decrypt`].
     ///
     /// Returns the session key.
-    fn decrypt_ciphertext(&mut self,
-                          ciphertext: &mpi::Ciphertext,
-                          plaintext_len: Option<usize>)
+    async fn decrypt_ciphertext(&mut self,
+                                ciphertext: &mpi::Ciphertext,
+                                plaintext_len: Option<usize>)
         -> Result<SessionKey>;
 
     /// Signs a message.
     ///
     /// `text` is the message to sign.
-    fn sign(&mut self, hash_algo: HashAlgorithm, text: &[u8])
+    async fn sign(&mut self, hash_algo: HashAlgorithm, text: &[u8])
         -> Result<(PublicKeyAlgorithm, mpi::Signature)>;
+
+    /// Exports the secret key material.
+    async fn export(&mut self)
+        -> Result<openpgp::packet::Key<
+            openpgp::packet::key::SecretParts,
+            openpgp::packet::key::UnspecifiedRole>>;
+
+    /// Changes the key's password.
+    ///
+    /// Changes the password.  Before calling this function, you
+    /// should call [`KeyHandle::password_source`] to determine if you
+    /// need to unlock the key, and whether you need to provide the
+    /// new password directly, or if that is obtained externally.
+    ///
+    /// If `new_password` is `Some` and [`KeyHandle::password_source`]
+    /// indicates that the password cannot be provided inline, then
+    /// the backend must return [`Error::NoInlinePassword`].  Likewise,
+    /// if `new_password` is `None`, and
+    /// [`KeyHandle::password_source`] does not indicate that the user
+    /// can be prompted for the password, then the backend must return
+    /// [`Error::NoExternalPassword`].
+    async fn change_password(&mut self, new_password: Option<&Password>)
+         -> Result<()>;
+
+    /// Deletes the key.
+    ///
+    /// This destroys the key's secret key material.
+    ///
+    /// If the key has to be unlocked, and the key is locked, the
+    /// backend should return an error.
+    ///
+    /// If the device managing the key does not support deleting keys,
+    /// then it should return [`Error::OperationNotSupported`].
+    async fn delete_secret_key_material(&mut self) -> Result<()>;
 }
