@@ -195,11 +195,16 @@ where B: Backend
     let mut ppr = PacketParser::from_bytes(ciphertext)?;
     while let PacketParserResult::Some(mut pp) = ppr {
         match pp.packet {
-            Packet::PKESK(ref pkesk) => {
+            Packet::PKESK(ref pkesk) if pkesk.recipient() == keyid => {
                 // Manual search.
+                log::debug!("Considering PKESK for {}", pkesk.recipient());
+
                 'done: for device in backend.list().await {
                     for mut key in device.list().await {
-                        if pkesk.recipient() != keyid {
+                        log::debug!("Considering key {}", key.fingerprint());
+
+                        if pkesk.recipient() != &key.keyid() {
+                            log::debug!("Wrong key for PKESK {}", key.fingerprint());
                             continue;
                         }
 
@@ -228,10 +233,16 @@ where B: Backend
                     }
                 }
             }
+            Packet::PKESK(ref pkesk) if pkesk.recipient() == keyid => {
+                // Manual search.
+                log::debug!("Ignoring PKESK for {}", pkesk.recipient());
+            }
             Packet::SEIP(_) => {
                 if let Some(session_key) = session_key.take() {
                     let (algo, sk) = session_key;
+                    log::debug!("Decrypting PKESK");
                     pp.decrypt(algo, &sk)?;
+                    log::debug!("Decrypting PKESK succeeded");
                 } else {
                     return Err(anyhow::anyhow!(
                         "Could not decrypt a PKESK".to_string()));
@@ -269,22 +280,99 @@ where B: Backend
     Ok(())
 }
 
+/// Preinit is a function that is called at the start of every
+/// test.  If it returns `true`, the test should be run.  If it
+/// returns `false`, then the test should be skipped.  This is useful
+/// when the required test infrastructure is not detected.
+///
+/// `$serialize_tests` is whether to serialize the tests for this
+/// backend.  If `false`, the tests are run concurrently.
+///
+/// `$import_cert` imports a certificate into the backend.
+///
+/// `$can_import_encrypted_secret_key_material` is whether
+/// `$import_cert` can import encrypted secret key material.
+///
+/// If `$key_sets` is `Some`, then the backend can only load standard
+/// keys consisting of (up to) an encrypt key, a signing key and an
+/// authentication key.  The integer indicates the number of key sets
+/// that the backend can load simultaneously.  If `$key_sets` is
+/// `None`, then the backend can load an infinite number of keys.
+///
+/// Some backends always protect keys by default.  If
+/// `$default_password` is `Some` that's the password for keys when
+/// they are imported.
 #[macro_export]
 macro_rules! generate_tests {
-    ($backend: ty, $init_backend: expr, $import_cert: expr) => {
+    ($preinit: expr, $serialize_tests: expr,
+     $backend: ty, $init_backend: expr,
+     $import_cert: expr, $can_import_encrypted_secret_key_material: expr,
+     $key_sets: expr, $default_password: expr,
+     $can_export: expr,
+     $can_change_password: expr,
+     $can_delete: expr) => {
+        static SERIALIZE: std::sync::Mutex<()> = std::sync::Mutex::new(());
+
+        fn serialize_maybe() -> Option<std::sync::MutexGuard<'static, ()>> {
+            let serialize_tests: bool = $serialize_tests;
+            if serialize_tests {
+                loop {
+                    match SERIALIZE.lock() {
+                        Ok(guard) => return Some(guard),
+                        Err(_err) => {
+                            // Another test panicked.  Don't let that
+                            // stop us.
+                            SERIALIZE.clear_poison();
+                        }
+                    }
+                }
+            } else {
+                None
+            }
+        }
+
+        /// strip the primary key's secret key material, if any.
+        ///
+        /// This is useful when `$key_sets` is `Some` and the backend
+        /// doesn't have space for the primary key.
+        fn strip_primary_secret(cert: Cert) -> Cert {
+            if cert.primary_key().has_secret() {
+                let mut pk = cert.primary_key().key().clone();
+                pk.steal_secret();
+
+                let cert = cert.insert_packets(pk).expect("can merge");
+                assert!(! cert.primary_key().has_secret());
+                cert
+            } else {
+                cert
+            }
+        }
+
         // Make sure $import_cert works.
         #[tokio::test]
         pub async fn import_cert_helper() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // A simple test: import a certificate.  Make sure the
             // keys are actually imported.
 
             let mut backend = $init_backend().await;
-            let cert = Cert::from_bytes(
+            let mut cert = Cert::from_bytes(
                 $crate::test_framework::data::key("no-password.asc"))
                 .expect("valid cert");
 
+            let key_sets: Option<usize> = $key_sets;
+            if key_sets.is_some() {
+                cert = strip_primary_secret(cert);
+            }
+
             $import_cert(&mut backend, &cert).await;
 
             let mut imported_keys = Vec::new();
@@ -326,6 +414,13 @@ macro_rules! generate_tests {
         pub async fn backend_import_cert() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // A simple test: import a certificate.  Make sure the
             // keys are actually imported.
 
@@ -414,9 +509,16 @@ macro_rules! generate_tests {
         }
 
         #[tokio::test]
-        pub async fn decrypt() -> Result<()> {
+        pub async fn decrypt_simple() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // A simple test: given a secret key and a message encrypted to
             // the key, can we use the backend to decrypt it?
 
@@ -429,18 +531,40 @@ macro_rules! generate_tests {
 
             let msg: &[u8] = $crate::test_framework::data::message("no-password.asc");
             let keyid: sequoia_openpgp::KeyID
-                = "4EEECCE4F09D9C92".parse().expect("valid");
+                = "AA5BFAC1E4A121A0".parse().expect("valid");
 
-            $crate::test_framework::try_decrypt(
-                &mut backend, msg, Some(b"foo\n"), &keyid, None).await.unwrap();
-
-            Ok(())
+            let default_password: Option<&str> = $default_password;
+            let default_password: Option<Password>
+                = default_password.map(|p| Password::from(p));
+            if let Err(err) = $crate::test_framework::try_decrypt(
+                &mut backend, msg, Some(b"foo\n"), &keyid, default_password).await
+            {
+                eprintln!("Decrypting: {}", err);
+                Err(err)
+            } else {
+                Ok(())
+            }
         }
 
         #[tokio::test]
         pub async fn decrypt_missing_password() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let can_import_encrypted_secret_key_material: bool
+                = $can_import_encrypted_secret_key_material;
+            if ! can_import_encrypted_secret_key_material {
+                log::debug!("Backend doesn't support importing encrypt secret \
+                             material.  Skipping test.");
+                return Ok(());
+            }
+
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // Load a few variants of a certificate: one where the
             // keys are encrypted with the password "bob", one where
             // they are encrypted with the password "smithy", and one
@@ -474,6 +598,21 @@ macro_rules! generate_tests {
         pub async fn decrypt_with_password() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let can_import_encrypted_secret_key_material: bool
+                = $can_import_encrypted_secret_key_material;
+            if ! can_import_encrypted_secret_key_material {
+                log::debug!("Backend doesn't support importing encrypt secret \
+                             material.  Skipping test.");
+                return Ok(());
+            }
+
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // Load a certificate where the keys are encrypted with
             // the password "bob".  Make sure we can decrypt the
             // message.  Load a variant of the certificate where the
@@ -488,7 +627,7 @@ macro_rules! generate_tests {
             let keyid: sequoia_openpgp::KeyID
                 = "0A4BA2249168D779".parse().expect("valid");
 
-            // Import Bob's certificate.  They keys are password
+            // Import Bob's certificate.  The keys are password
             // protected.  Make sure we can use them with the right
             // password, and can't with the wrong one.
             log::info!("Importing certificate variant #1");
@@ -555,8 +694,15 @@ macro_rules! generate_tests {
         pub async fn verify() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             async fn test(filename: &str, signing_key: &str,
-                          password: Option<Password>,
+                          password: Option<&str>,
                           success: bool)
                 -> Result<()>
             {
@@ -568,10 +714,19 @@ macro_rules! generate_tests {
                 $import_cert(&mut backend, &cert).await;
 
                 let mut signer = backend.find_key(&signing_key).await?;
-                if let Some(password) = password.as_ref() {
-                    if let Err(err) = signer.unlock(Some(password)).await {
+                if let Some(password) = password {
+                    if let Err(err) = signer.unlock(Some(&password.into())).await {
                         log::warn!("Failed to unlock key: {}", err);
                     }
+                } else {
+                    // Use the default password (if configured).
+                    let default_password: Option<&str> = $default_password;
+                    if let Some(default_password) = default_password {
+                        let default_password = Password::from(default_password);
+                        if let Err(err) = signer.unlock(Some(&default_password)).await {
+                            log::warn!("Failed to unlock key: {}", err);
+                        }
+                    }
                 }
 
                 let signing_fpr = Fingerprint::from_hex(signing_key)
@@ -594,9 +749,17 @@ macro_rules! generate_tests {
             }
 
             // A simple test: sign a message using a key with no password.
-            test("no-password.asc", "88647F753426C4F16D1ED1E50EF5A4511075ACD7",
+            test("no-password.asc", "8D46B2975890A615AF4109CB6BA1BF391DF802C8",
                  None, true).await.expect("should pass");
 
+            let can_import_encrypted_secret_key_material: bool
+                = $can_import_encrypted_secret_key_material;
+            if ! can_import_encrypted_secret_key_material {
+                log::debug!("Backend doesn't support importing encrypt secret \
+                             material.  Skipping the rest of the test.");
+                return Ok(());
+            }
+
             // Sign a message using a key that is password protected,
             // but don't provide the password.  Make sure it fails.
             test("bob-password-bob.asc", "9410E96E68643821794608269CB5006116833EE7",
@@ -604,11 +767,11 @@ macro_rules! generate_tests {
 
             // Now provide the password.
             test("bob-password-bob.asc", "9410E96E68643821794608269CB5006116833EE7",
-                 Some("bob".into()), true).await.expect("should pass");
+                 Some("bob"), true).await.expect("should pass");
 
             // And now provide the wrong password.
             test("bob-password-bob.asc", "9410E96E68643821794608269CB5006116833EE7",
-                 Some("wrong password".into()), false).await.expect("should pass");
+                 Some("wrong password"), false).await.expect("should pass");
 
             Ok(())
         }
@@ -617,53 +780,99 @@ macro_rules! generate_tests {
         pub async fn list_keys() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
+            let key_sets: Option<usize> = $key_sets;
+
+            let _ = env_logger::Builder::from_default_env().try_init();
+
+            // We first make sure that we can list the empty backend.
+            {
+                let mut backend = init_backend().await;
+
+                let mut got = Vec::new();
+                for device in backend.list().await {
+                    for mut key in device.list().await {
+                        got.push(key.fingerprint().to_string());
+                    }
+                }
+                assert!(got.is_empty());
+            }
+
             // We have two variants of a certificate with disjoint
             // subkeys.  When we load both certificates, make sure both
             // sets of variants appear.
 
             let keyring_a: &[u8] = $crate::test_framework::data::key(
                 "carol-subkeys-a.asc");
-            let keys_a = &[
-                "B0B75BE1004E71FC9FEEF319CC160DCFDE911D68",
-                "4BA88DAF3998EF1A1A4DBC9349F15C0BB3DD62B5",
-            ];
+            let mut cert_a = Cert::from_bytes(keyring_a)
+                .expect("valid cert");
+            if key_sets.is_some() {
+                // Strip the primary.
+                cert_a = strip_primary_secret(cert_a);
+            }
+            let keys_a = cert_a.keys().secret().map(|k| k.fingerprint())
+                .collect::<Vec<Fingerprint>>();
 
             let keyring_b: &[u8] = $crate::test_framework::data::key(
                 "carol-subkeys-b.asc");
-            let keys_b = &[
-                "B0B75BE1004E71FC9FEEF319CC160DCFDE911D68",
-                "61D8031424FC2C58BBA6CA7F11B1E2EDC6E4D1D1",
-                "5606791C6C58CD2A8F1306397189B15EF88BB1D5",
-            ];
+            let mut cert_b = Cert::from_bytes(keyring_b)
+                .expect("valid cert");
+            if key_sets.is_some() {
+                // Strip the primary.
+                cert_b = strip_primary_secret(cert_b);
+            }
+            let keys_b = cert_b.keys().secret().map(|k| k.fingerprint())
+                .collect::<Vec<Fingerprint>>();
+
+            // We expect the same certificate, but with different
+            // subkeys.
+            assert_eq!(cert_a.fingerprint(), cert_b.fingerprint());
+            assert_ne!(keys_a, keys_b);
+
+            let mut all_keys = keys_a.clone();
+            all_keys.extend_from_slice(&keys_b);
+            all_keys.sort();
+            all_keys.dedup();
 
             // Try with just a, then just b, then both variants.
             for (a, b) in [(true, false), (false, true), (true, true)].iter() {
-                let mut keyring: Vec<&[u8]> = Vec::new();
-                let mut keys = Vec::new();
+                if *a && *b {
+                    eprintln!("Loading a and b: {:?} {:?}", keys_a, keys_b);
+                } else if *a {
+                    eprintln!("Loading a: {:?}", keys_a);
+                } else if *b {
+                    eprintln!("Loading b: {:?}", keys_b);
+                }
+
+                let mut keyring: Vec<&Cert> = Vec::new();
+                let mut keys: Vec<Fingerprint> = Vec::new();
 
                 if *a {
-                    keyring.push(keyring_a);
-                    keys.extend_from_slice(keys_a);
+                    keyring.push(&cert_a);
+                    keys.extend_from_slice(&keys_a);
                 }
                 if *b {
-                    keyring.push(keyring_b);
-                    keys.extend_from_slice(keys_b);
+                    keyring.push(&cert_b);
+                    keys.extend_from_slice(&keys_b);
                 }
 
                 let mut backend = init_backend().await;
 
-                for cert_bytes in keyring {
-                    let cert = Cert::from_bytes(cert_bytes)
-                        .expect("valid cert");
-
-                    $import_cert(&mut backend, &cert).await;
+                for cert in keyring {
+                    $import_cert(&mut backend, cert).await;
                 }
 
                 // Check that we can find the cert by iterating.
                 let mut got = Vec::new();
                 for device in backend.list().await {
                     for mut key in device.list().await {
-                        got.push(key.fingerprint().to_string());
+                        got.push(key.fingerprint());
                     }
                 }
                 got.sort();
@@ -671,23 +880,27 @@ macro_rules! generate_tests {
                 keys.sort();
                 keys.dedup();
 
-                assert_eq!(
-                    keys.into_iter().map(String::from).collect::<Vec<String>>(),
-                    got);
+                assert_eq!(keys, got,
+                    "Expected: {:?}\nGot: {:?}", keys, got);
 
                 // Check that Backend::find_key returns all of the keys.
-                for (i, k) in keys_a.iter().enumerate() {
-                    if *a || i == 0 {
-                        assert!(backend.find_key(k).await.is_ok());
-                    } else {
-                        assert!(backend.find_key(k).await.is_err());
-                    }
-                }
-                for (i, k) in keys_b.iter().enumerate() {
-                    if *b || i == 0 {
-                        assert!(backend.find_key(k).await.is_ok());
-                    } else {
-                        assert!(backend.find_key(k).await.is_err());
+                for k in all_keys.iter() {
+                    let expected = keys.contains(k);
+                    eprintln!("Checking find_key({}) => {}",
+                              k.to_string(), expected);
+                    match (expected, backend.find_key(&k.to_string()).await) {
+                        (true, Ok(_)) => {
+                        }
+                        (true, Err(err)) => {
+                            panic!("Didn't find {}, but should have. \
+                                    Error message is: {}",
+                                   k, err);
+                        }
+                        (false, Ok(_)) => {
+                            panic!("Found {}, but shouldn't have", k);
+                        }
+                        (false, Err(_err)) => {
+                        }
                     }
                 }
             }
@@ -699,48 +912,55 @@ macro_rules! generate_tests {
         pub async fn list_only_returns_secret_keys() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // Make sure only keys with secret key material are returned.
 
             let cert_bytes: &[u8] = test_framework::data::key(
                 "dave-not-all-keys-have-secrets.asc");
-            let cert = "4FD198623B8726286C7159A6032CCEF246B60258";
-            let mut keys = vec![
-                "AC32461C491647DA6DAC794712CCFC305567FA9D",
-            ];
-            let public_keys = vec![
-                "4FD198623B8726286C7159A6032CCEF246B60258",
-                "AA91F4FA667D5B17E8ACCA9F6BC89BD62B55B802",
-            ];
+            let cert = Cert::from_bytes(cert_bytes)
+                .expect("valid cert");
+
+            let (keys, public_keys): (Vec<(Fingerprint, bool)> , _) = cert
+                .keys()
+                .map(|k| (k.fingerprint(), k.has_secret()))
+                .partition(|(_fpr, has_secret)| *has_secret);
+
+            let keys = keys.into_iter().map(|(fpr, _)| fpr)
+                .collect::<Vec<Fingerprint>>();
+            let public_keys = public_keys.into_iter().map(|(fpr, _)| fpr)
+                .collect::<Vec<Fingerprint>>();
+
+            assert!(! keys.is_empty());
+            assert!(! public_keys.is_empty());
 
             let mut backend = init_backend().await;
 
-            let cert = Cert::from_bytes(cert_bytes)
-                .expect("valid cert");
             $import_cert(&mut backend, &cert).await;
 
             // List the keys.
             let mut got = Vec::new();
             for device in backend.list().await {
                 for key in device.list().await {
-                    got.push(key.id());
+                    got.push(key.fingerprint());
                 }
             }
             got.sort();
 
-            keys.sort();
-            keys.dedup();
-
-            assert_eq!(
-                keys.iter().map(|&k| k.into()).collect::<Vec<String>>(),
-                got);
+            assert_eq!(keys, got);
 
             // Check that Backend::find_key returns all of the keys.
-            for k in keys {
-                assert!(backend.find_key(k).await.is_ok());
+            for k in keys.iter() {
+                assert!(backend.find_key(&k.to_string()).await.is_ok());
             }
             // And doesn't return the public keys.
-            for k in public_keys {
-                assert!(backend.find_key(k).await.is_err());
+            for k in public_keys.iter() {
+                assert!(backend.find_key(&k.to_string()).await.is_err());
             }
 
             Ok(())
@@ -751,6 +971,20 @@ macro_rules! generate_tests {
         pub async fn key_export() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let can_export: bool = $can_export;
+            if ! can_export {
+                log::debug!("Backend doesn't support exporting secret key \
+                             material.  Skipping test.");
+                return Ok(());
+            }
+
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // A simple test: import a certificate.  Make sure the
             // keys are actually imported.
 
@@ -793,6 +1027,20 @@ macro_rules! generate_tests {
         pub async fn change_password() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let can_change_password: bool = $can_change_password;
+            if ! can_change_password {
+                log::debug!("Backend doesn't support changing passwords. \
+                             Skipping test.");
+                return Ok(());
+            }
+
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // Load a certificate where the keys are encrypted with
             // the password "bob".  Make sure we can use it to decrypt
             // a message.  Change the key's password, and relock it.
@@ -869,6 +1117,20 @@ macro_rules! generate_tests {
         pub async fn delete_secret_key_material() -> Result<()> {
             let _ = env_logger::Builder::from_default_env().try_init();
 
+            let can_delete: bool = $can_delete;
+            if ! can_delete {
+                log::debug!("Backend doesn't support deleting secret key \
+                             material.  Skipping test.");
+                return Ok(());
+            }
+
+            let _serialized = serialize_maybe();
+
+            if ! $preinit() {
+                eprintln!("preinit returned false, skipping test.");
+                return Ok(());
+            }
+
             // Load a certificate where the keys are encrypted with
             // the password "bob".  Make sure we can use it to decrypt
             // a message.  Delete the primary key.  Make sure we can
